@@ -1,13 +1,14 @@
 import ssl
 import sys
+import time
 from types import TracebackType
-from typing import AsyncIterable, AsyncIterator, List, Optional, Type
+from typing import AsyncIterable, AsyncIterator, Iterable, List, Optional, Type
 
-from .._exceptions import ConnectionNotAvailable, UnsupportedProtocol
+from .._backends.auto import AutoBackend
+from .._backends.base import SOCKET_OPTION, AsyncNetworkBackend
+from .._exceptions import ConnectionNotAvailable, PoolTimeout, UnsupportedProtocol
 from .._models import Origin, Request, Response
-from .._synchronization import AsyncEvent, AsyncLock
-from ..backends.auto import AutoBackend
-from ..backends.base import AsyncNetworkBackend
+from .._synchronization import AsyncEvent, AsyncLock, AsyncShieldCancellation
 from .connection import AsyncHTTPConnection
 from .interfaces import AsyncConnectionInterface, AsyncRequestInterface
 
@@ -54,6 +55,7 @@ class AsyncConnectionPool(AsyncRequestInterface):
         local_address: Optional[str] = None,
         uds: Optional[str] = None,
         network_backend: Optional[AsyncNetworkBackend] = None,
+        socket_options: Optional[Iterable[SOCKET_OPTION]] = None,
     ) -> None:
         """
         A connection pool for making HTTP requests.
@@ -81,6 +83,8 @@ class AsyncConnectionPool(AsyncRequestInterface):
                 `local_address="::"` will connect using an `AF_INET6` address (IPv6).
             uds: Path to a Unix Domain Socket to use instead of TCP sockets.
             network_backend: A backend instance to use for handling network I/O.
+            socket_options: Socket options that have to be included
+             in the TCP socket when the connection was established.
         """
         self._ssl_context = ssl_context
 
@@ -109,6 +113,7 @@ class AsyncConnectionPool(AsyncRequestInterface):
         self._network_backend = (
             AutoBackend() if network_backend is None else network_backend
         )
+        self._socket_options = socket_options
 
     def create_connection(self, origin: Origin) -> AsyncConnectionInterface:
         return AsyncHTTPConnection(
@@ -121,6 +126,7 @@ class AsyncConnectionPool(AsyncRequestInterface):
             local_address=self._local_address,
             uds=self._uds,
             network_backend=self._network_backend,
+            socket_options=self._socket_options,
         )
 
     @property
@@ -215,6 +221,13 @@ class AsyncConnectionPool(AsyncRequestInterface):
             )
 
         status = RequestStatus(request)
+        timeouts = request.extensions.get("timeout", {})
+        timeout = timeouts.get("pool", None)
+
+        if timeout is not None:
+            deadline = time.monotonic() + timeout
+        else:
+            deadline = float("inf")
 
         async with self._pool_lock:
             self._requests.append(status)
@@ -222,8 +235,6 @@ class AsyncConnectionPool(AsyncRequestInterface):
             await self._attempt_to_acquire_connection(status)
 
         while True:
-            timeouts = request.extensions.get("timeout", {})
-            timeout = timeouts.get("pool", None)
             try:
                 connection = await status.wait_for_connection(timeout=timeout)
             except BaseException as exc:
@@ -252,11 +263,16 @@ class AsyncConnectionPool(AsyncRequestInterface):
                     status.unset_connection()
                     await self._attempt_to_acquire_connection(status)
             except BaseException as exc:
-                await self.response_closed(status)
+                with AsyncShieldCancellation():
+                    await self.response_closed(status)
                 raise exc
             else:
                 break
 
+            timeout = deadline - time.monotonic()
+            if timeout < 0:
+                raise PoolTimeout  # pragma: nocover
+
         # When we return the response, we wrap the stream in a special class
         # that handles notifying the connection pool once the response
         # has been released.
@@ -310,6 +326,10 @@ class AsyncConnectionPool(AsyncRequestInterface):
             self._requests = []
 
     async def __aenter__(self) -> "AsyncConnectionPool":
+        # Acquiring the pool lock here ensures that we have the
+        # correct dependencies installed as early as possible.
+        async with self._pool_lock:
+            pass
         return self
 
     async def __aexit__(
@@ -346,4 +366,5 @@ class ConnectionPoolByteStream:
             if hasattr(self._stream, "aclose"):
                 await self._stream.aclose()
         finally:
-            await self._pool.response_closed(self._status)
+            with AsyncShieldCancellation():
+                await self._pool.response_closed(self._status)
