@@ -1,18 +1,12 @@
 import logging
-from typing import List, Optional
+import typing
 
+import hpack
+import hyperframe.frame
 import pytest
 from tests import concurrency
 
-from httpcore import (
-    ConnectionPool,
-    ConnectError,
-    PoolTimeout,
-    ReadError,
-    UnsupportedProtocol,
-)
-from httpcore.backends.base import NetworkStream
-from httpcore.backends.mock import MockBackend
+import httpcore
 
 
 
@@ -20,7 +14,7 @@ def test_connection_pool_with_keepalive():
     """
     By default HTTP/1.1 requests should be returned to the connection pool.
     """
-    network_backend = MockBackend(
+    network_backend = httpcore.MockBackend(
         [
             b"HTTP/1.1 200 OK\r\n",
             b"Content-Type: plain/text\r\n",
@@ -35,7 +29,7 @@ def test_connection_pool_with_keepalive():
         ]
     )
 
-    with ConnectionPool(
+    with httpcore.ConnectionPool(
         network_backend=network_backend,
     ) as pool:
         # Sending an intial request, which once complete will return to the pool, IDLE.
@@ -92,7 +86,7 @@ def test_connection_pool_with_close():
     HTTP/1.1 requests that include a 'Connection: Close' header should
     not be returned to the connection pool.
     """
-    network_backend = MockBackend(
+    network_backend = httpcore.MockBackend(
         [
             b"HTTP/1.1 200 OK\r\n",
             b"Content-Type: plain/text\r\n",
@@ -102,7 +96,7 @@ def test_connection_pool_with_close():
         ]
     )
 
-    with ConnectionPool(network_backend=network_backend) as pool:
+    with httpcore.ConnectionPool(network_backend=network_backend) as pool:
         # Sending an intial request, which once complete will not return to the pool.
         with pool.stream(
             "GET", "https://example.com/", headers={"Connection": "close"}
@@ -120,12 +114,129 @@ def test_connection_pool_with_close():
 
 
 
+def test_connection_pool_with_http2():
+    """
+    Test a connection pool with HTTP/2 requests.
+    """
+    network_backend = httpcore.MockBackend(
+        buffer=[
+            hyperframe.frame.SettingsFrame().serialize(),
+            hyperframe.frame.HeadersFrame(
+                stream_id=1,
+                data=hpack.Encoder().encode(
+                    [
+                        (b":status", b"200"),
+                        (b"content-type", b"plain/text"),
+                    ]
+                ),
+                flags=["END_HEADERS"],
+            ).serialize(),
+            hyperframe.frame.DataFrame(
+                stream_id=1, data=b"Hello, world!", flags=["END_STREAM"]
+            ).serialize(),
+            hyperframe.frame.HeadersFrame(
+                stream_id=3,
+                data=hpack.Encoder().encode(
+                    [
+                        (b":status", b"200"),
+                        (b"content-type", b"plain/text"),
+                    ]
+                ),
+                flags=["END_HEADERS"],
+            ).serialize(),
+            hyperframe.frame.DataFrame(
+                stream_id=3, data=b"Hello, world!", flags=["END_STREAM"]
+            ).serialize(),
+        ],
+        http2=True,
+    )
+
+    with httpcore.ConnectionPool(
+        network_backend=network_backend,
+    ) as pool:
+        # Sending an intial request, which once complete will return to the pool, IDLE.
+        response = pool.request("GET", "https://example.com/")
+        assert response.status == 200
+        assert response.content == b"Hello, world!"
+
+        info = [repr(c) for c in pool.connections]
+        assert info == [
+            "<HTTPConnection ['https://example.com:443', HTTP/2, IDLE, Request Count: 1]>"
+        ]
+
+        # Sending a second request to the same origin will reuse the existing IDLE connection.
+        response = pool.request("GET", "https://example.com/")
+        assert response.status == 200
+        assert response.content == b"Hello, world!"
+
+        info = [repr(c) for c in pool.connections]
+        assert info == [
+            "<HTTPConnection ['https://example.com:443', HTTP/2, IDLE, Request Count: 2]>"
+        ]
+
+
+
+def test_connection_pool_with_http2_goaway():
+    """
+    Test a connection pool with HTTP/2 requests, that cleanly disconnects
+    with a GoAway frame after the first request.
+    """
+    network_backend = httpcore.MockBackend(
+        buffer=[
+            hyperframe.frame.SettingsFrame().serialize(),
+            hyperframe.frame.HeadersFrame(
+                stream_id=1,
+                data=hpack.Encoder().encode(
+                    [
+                        (b":status", b"200"),
+                        (b"content-type", b"plain/text"),
+                    ]
+                ),
+                flags=["END_HEADERS"],
+            ).serialize(),
+            hyperframe.frame.DataFrame(
+                stream_id=1, data=b"Hello, world!", flags=["END_STREAM"]
+            ).serialize(),
+            hyperframe.frame.GoAwayFrame(
+                stream_id=0, error_code=0, last_stream_id=1
+            ).serialize(),
+            b"",
+        ],
+        http2=True,
+    )
+
+    with httpcore.ConnectionPool(
+        network_backend=network_backend,
+    ) as pool:
+        # Sending an intial request, which once complete will return to the pool, IDLE.
+        response = pool.request("GET", "https://example.com/")
+        assert response.status == 200
+        assert response.content == b"Hello, world!"
+
+        info = [repr(c) for c in pool.connections]
+        assert info == [
+            "<HTTPConnection ['https://example.com:443', HTTP/2, IDLE, Request Count: 1]>"
+        ]
+
+        # Sending a second request to the same origin will require a new connection.
+        response = pool.request("GET", "https://example.com/")
+        assert response.status == 200
+        assert response.content == b"Hello, world!"
+
+        info = [repr(c) for c in pool.connections]
+        assert info == [
+            "<HTTPConnection ['https://example.com:443', HTTP/2, IDLE, Request Count: 1]>",
+            "<HTTPConnection ['https://example.com:443', HTTP/2, CLOSED, Request Count: 1]>",
+        ]
+
+
+
 def test_trace_request():
     """
     The 'trace' request extension allows for a callback function to inspect the
     internal events that occur while sending a request.
     """
-    network_backend = MockBackend(
+    network_backend = httpcore.MockBackend(
         [
             b"HTTP/1.1 200 OK\r\n",
             b"Content-Type: plain/text\r\n",
@@ -140,7 +251,7 @@ def test_trace_request():
     def trace(name, kwargs):
         called.append(name)
 
-    with ConnectionPool(network_backend=network_backend) as pool:
+    with httpcore.ConnectionPool(network_backend=network_backend) as pool:
         pool.request("GET", "https://example.com/", extensions={"trace": trace})
 
     assert called == [
@@ -169,7 +280,7 @@ def test_debug_request(caplog):
     """
     caplog.set_level(logging.DEBUG)
 
-    network_backend = MockBackend(
+    network_backend = httpcore.MockBackend(
         [
             b"HTTP/1.1 200 OK\r\n",
             b"Content-Type: plain/text\r\n",
@@ -179,53 +290,53 @@ def test_debug_request(caplog):
         ]
     )
 
-    with ConnectionPool(network_backend=network_backend) as pool:
+    with httpcore.ConnectionPool(network_backend=network_backend) as pool:
         pool.request("GET", "http://example.com/")
 
     assert caplog.record_tuples == [
         (
-            "httpcore",
+            "httpcore.connection",
             logging.DEBUG,
-            "connection.connect_tcp.started host='example.com' port=80 local_address=None timeout=None",
+            "connect_tcp.started host='example.com' port=80 local_address=None timeout=None socket_options=None",
         ),
         (
-            "httpcore",
+            "httpcore.connection",
             logging.DEBUG,
-            "connection.connect_tcp.complete return_value=<httpcore.MockStream>",
+            "connect_tcp.complete return_value=<httpcore.MockStream>",
         ),
         (
-            "httpcore",
+            "httpcore.http11",
             logging.DEBUG,
-            "http11.send_request_headers.started request=<Request [b'GET']>",
+            "send_request_headers.started request=<Request [b'GET']>",
         ),
-        ("httpcore", logging.DEBUG, "http11.send_request_headers.complete"),
+        ("httpcore.http11", logging.DEBUG, "send_request_headers.complete"),
         (
-            "httpcore",
+            "httpcore.http11",
             logging.DEBUG,
-            "http11.send_request_body.started request=<Request [b'GET']>",
+            "send_request_body.started request=<Request [b'GET']>",
         ),
-        ("httpcore", logging.DEBUG, "http11.send_request_body.complete"),
+        ("httpcore.http11", logging.DEBUG, "send_request_body.complete"),
         (
-            "httpcore",
+            "httpcore.http11",
             logging.DEBUG,
-            "http11.receive_response_headers.started request=<Request [b'GET']>",
+            "receive_response_headers.started request=<Request [b'GET']>",
         ),
         (
-            "httpcore",
+            "httpcore.http11",
             logging.DEBUG,
-            "http11.receive_response_headers.complete return_value="
+            "receive_response_headers.complete return_value="
             "(b'HTTP/1.1', 200, b'OK', [(b'Content-Type', b'plain/text'), (b'Content-Length', b'13')])",
         ),
         (
-            "httpcore",
+            "httpcore.http11",
             logging.DEBUG,
-            "http11.receive_response_body.started request=<Request [b'GET']>",
+            "receive_response_body.started request=<Request [b'GET']>",
         ),
-        ("httpcore", logging.DEBUG, "http11.receive_response_body.complete"),
-        ("httpcore", logging.DEBUG, "http11.response_closed.started"),
-        ("httpcore", logging.DEBUG, "http11.response_closed.complete"),
-        ("httpcore", logging.DEBUG, "connection.close.started"),
-        ("httpcore", logging.DEBUG, "connection.close.complete"),
+        ("httpcore.http11", logging.DEBUG, "receive_response_body.complete"),
+        ("httpcore.http11", logging.DEBUG, "response_closed.started"),
+        ("httpcore.http11", logging.DEBUG, "response_closed.complete"),
+        ("httpcore.connection", logging.DEBUG, "close.started"),
+        ("httpcore.connection", logging.DEBUG, "close.complete"),
     ]
 
 
@@ -235,14 +346,14 @@ def test_connection_pool_with_http_exception():
     HTTP/1.1 requests that result in an exception during the connection should
     not be returned to the connection pool.
     """
-    network_backend = MockBackend([b"Wait, this isn't valid HTTP!"])
+    network_backend = httpcore.MockBackend([b"Wait, this isn't valid HTTP!"])
 
     called = []
 
     def trace(name, kwargs):
         called.append(name)
 
-    with ConnectionPool(network_backend=network_backend) as pool:
+    with httpcore.ConnectionPool(network_backend=network_backend) as pool:
         # Sending an initial request, which once complete will not return to the pool.
         with pytest.raises(Exception):
             pool.request(
@@ -275,15 +386,18 @@ def test_connection_pool_with_connect_exception():
     be returned to the connection pool.
     """
 
-    class FailedConnectBackend(MockBackend):
+    class FailedConnectBackend(httpcore.MockBackend):
         def connect_tcp(
             self,
             host: str,
             port: int,
-            timeout: Optional[float] = None,
-            local_address: Optional[str] = None,
-        ) -> NetworkStream:
-            raise ConnectError("Could not connect")
+            timeout: typing.Optional[float] = None,
+            local_address: typing.Optional[str] = None,
+            socket_options: typing.Optional[
+                typing.Iterable[httpcore.SOCKET_OPTION]
+            ] = None,
+        ) -> httpcore.NetworkStream:
+            raise httpcore.ConnectError("Could not connect")
 
     network_backend = FailedConnectBackend([])
 
@@ -292,7 +406,7 @@ def test_connection_pool_with_connect_exception():
     def trace(name, kwargs):
         called.append(name)
 
-    with ConnectionPool(network_backend=network_backend) as pool:
+    with httpcore.ConnectionPool(network_backend=network_backend) as pool:
         # Sending an initial request, which once complete will not return to the pool.
         with pytest.raises(Exception):
             pool.request(
@@ -314,7 +428,7 @@ def test_connection_pool_with_immediate_expiry():
     Connection pools with keepalive_expiry=0.0 should immediately expire
     keep alive connections.
     """
-    network_backend = MockBackend(
+    network_backend = httpcore.MockBackend(
         [
             b"HTTP/1.1 200 OK\r\n",
             b"Content-Type: plain/text\r\n",
@@ -324,7 +438,7 @@ def test_connection_pool_with_immediate_expiry():
         ]
     )
 
-    with ConnectionPool(
+    with httpcore.ConnectionPool(
         keepalive_expiry=0.0,
         network_backend=network_backend,
     ) as pool:
@@ -348,7 +462,7 @@ def test_connection_pool_with_no_keepalive_connections_allowed():
     When 'max_keepalive_connections=0' is used, IDLE connections should not
     be returned to the pool.
     """
-    network_backend = MockBackend(
+    network_backend = httpcore.MockBackend(
         [
             b"HTTP/1.1 200 OK\r\n",
             b"Content-Type: plain/text\r\n",
@@ -358,7 +472,7 @@ def test_connection_pool_with_no_keepalive_connections_allowed():
         ]
     )
 
-    with ConnectionPool(
+    with httpcore.ConnectionPool(
         max_keepalive_connections=0, network_backend=network_backend
     ) as pool:
         # Sending an intial request, which once complete will not return to the pool.
@@ -381,7 +495,7 @@ def test_connection_pool_concurrency():
     HTTP/1.1 requests made in concurrency must not ever exceed the maximum number
     of allowable connection in the pool.
     """
-    network_backend = MockBackend(
+    network_backend = httpcore.MockBackend(
         [
             b"HTTP/1.1 200 OK\r\n",
             b"Content-Type: plain/text\r\n",
@@ -397,10 +511,10 @@ def test_connection_pool_concurrency():
             info_list.append(info)
             response.read()
 
-    with ConnectionPool(
+    with httpcore.ConnectionPool(
         max_connections=1, network_backend=network_backend
     ) as pool:
-        info_list: List[str] = []
+        info_list: typing.List[str] = []
         with concurrency.open_nursery() as nursery:
             for domain in ["a.com", "b.com", "c.com", "d.com", "e.com"]:
                 nursery.start_soon(fetch, pool, domain, info_list)
@@ -426,7 +540,7 @@ def test_connection_pool_concurrency_same_domain_closing():
     HTTP/1.1 requests made in concurrency must not ever exceed the maximum number
     of allowable connection in the pool.
     """
-    network_backend = MockBackend(
+    network_backend = httpcore.MockBackend(
         [
             b"HTTP/1.1 200 OK\r\n",
             b"Content-Type: plain/text\r\n",
@@ -443,10 +557,10 @@ def test_connection_pool_concurrency_same_domain_closing():
             info_list.append(info)
             response.read()
 
-    with ConnectionPool(
+    with httpcore.ConnectionPool(
         max_connections=1, network_backend=network_backend, http2=True
     ) as pool:
-        info_list: List[str] = []
+        info_list: typing.List[str] = []
         with concurrency.open_nursery() as nursery:
             for domain in ["a.com", "a.com", "a.com", "a.com", "a.com"]:
                 nursery.start_soon(fetch, pool, domain, info_list)
@@ -468,7 +582,7 @@ def test_connection_pool_concurrency_same_domain_keepalive():
     HTTP/1.1 requests made in concurrency must not ever exceed the maximum number
     of allowable connection in the pool.
     """
-    network_backend = MockBackend(
+    network_backend = httpcore.MockBackend(
         [
             b"HTTP/1.1 200 OK\r\n",
             b"Content-Type: plain/text\r\n",
@@ -485,10 +599,10 @@ def test_connection_pool_concurrency_same_domain_keepalive():
             info_list.append(info)
             response.read()
 
-    with ConnectionPool(
+    with httpcore.ConnectionPool(
         max_connections=1, network_backend=network_backend, http2=True
     ) as pool:
-        info_list: List[str] = []
+        info_list: typing.List[str] = []
         with concurrency.open_nursery() as nursery:
             for domain in ["a.com", "a.com", "a.com", "a.com", "a.com"]:
                 nursery.start_soon(fetch, pool, domain, info_list)
@@ -509,11 +623,11 @@ def test_connection_pool_concurrency_same_domain_keepalive():
 
 
 def test_unsupported_protocol():
-    with ConnectionPool() as pool:
-        with pytest.raises(UnsupportedProtocol):
+    with httpcore.ConnectionPool() as pool:
+        with pytest.raises(httpcore.UnsupportedProtocol):
             pool.request("GET", "ftp://www.example.com/")
 
-        with pytest.raises(UnsupportedProtocol):
+        with pytest.raises(httpcore.UnsupportedProtocol):
             pool.request("GET", "://www.example.com/")
 
 
@@ -523,7 +637,7 @@ def test_connection_pool_closed_while_request_in_flight():
     Closing a connection pool while a request/response is still in-flight
     should raise an error.
     """
-    network_backend = MockBackend(
+    network_backend = httpcore.MockBackend(
         [
             b"HTTP/1.1 200 OK\r\n",
             b"Content-Type: plain/text\r\n",
@@ -533,14 +647,14 @@ def test_connection_pool_closed_while_request_in_flight():
         ]
     )
 
-    with ConnectionPool(
+    with httpcore.ConnectionPool(
         network_backend=network_backend,
     ) as pool:
         # Send a request, and then close the connection pool while the
         # response has not yet been streamed.
         with pool.stream("GET", "https://example.com/") as response:
             pool.close()
-            with pytest.raises(ReadError):
+            with pytest.raises(httpcore.ReadError):
                 response.read()
 
 
@@ -549,7 +663,7 @@ def test_connection_pool_timeout():
     """
     Ensure that exceeding max_connections can cause a request to timeout.
     """
-    network_backend = MockBackend(
+    network_backend = httpcore.MockBackend(
         [
             b"HTTP/1.1 200 OK\r\n",
             b"Content-Type: plain/text\r\n",
@@ -559,19 +673,81 @@ def test_connection_pool_timeout():
         ]
     )
 
-    with ConnectionPool(
+    with httpcore.ConnectionPool(
         network_backend=network_backend, max_connections=1
     ) as pool:
         # Send a request to a pool that is configured to only support a single
         # connection, and then ensure that a second concurrent request
         # fails with a timeout.
         with pool.stream("GET", "https://example.com/"):
-            with pytest.raises(PoolTimeout):
+            with pytest.raises(httpcore.PoolTimeout):
                 extensions = {"timeout": {"pool": 0.0001}}
                 pool.request("GET", "https://example.com/", extensions=extensions)
 
 
 
+def test_connection_pool_timeout_zero():
+    """
+    A pool timeout of 0 shouldn't raise a PoolTimeout if there's
+    no need to wait on a new connection.
+    """
+    network_backend = httpcore.MockBackend(
+        [
+            b"HTTP/1.1 200 OK\r\n",
+            b"Content-Type: plain/text\r\n",
+            b"Content-Length: 13\r\n",
+            b"\r\n",
+            b"Hello, world!",
+            b"HTTP/1.1 200 OK\r\n",
+            b"Content-Type: plain/text\r\n",
+            b"Content-Length: 13\r\n",
+            b"\r\n",
+            b"Hello, world!",
+        ]
+    )
+
+    # Use a pool timeout of zero.
+    extensions = {"timeout": {"pool": 0}}
+
+    # A connection pool configured to allow only one connection at a time.
+    with httpcore.ConnectionPool(
+        network_backend=network_backend, max_connections=1
+    ) as pool:
+        # Two consecutive requests with a pool timeout of zero.
+        # Both succeed without raising a timeout.
+        response = pool.request(
+            "GET", "https://example.com/", extensions=extensions
+        )
+        assert response.status == 200
+        assert response.content == b"Hello, world!"
+
+        response = pool.request(
+            "GET", "https://example.com/", extensions=extensions
+        )
+        assert response.status == 200
+        assert response.content == b"Hello, world!"
+
+    # A connection pool configured to allow only one connection at a time.
+    with httpcore.ConnectionPool(
+        network_backend=network_backend, max_connections=1
+    ) as pool:
+        # Two concurrent requests with a pool timeout of zero.
+        # Only the first will succeed without raising a timeout.
+        with pool.stream(
+            "GET", "https://example.com/", extensions=extensions
+        ) as response:
+            # The first response hasn't yet completed.
+            with pytest.raises(httpcore.PoolTimeout):
+                # So a pool timeout occurs.
+                pool.request("GET", "https://example.com/", extensions=extensions)
+            # The first response now completes.
+            response.read()
+
+        assert response.status == 200
+        assert response.content == b"Hello, world!"
+
+
+
 def test_http11_upgrade_connection():
     """
     HTTP "101 Switching Protocols" indicates an upgraded connection.
@@ -582,7 +758,7 @@ def test_http11_upgrade_connection():
     https://httpwg.org/specs/rfc9110.html#status.101
     https://developer.mozilla.org/en-US/docs/Web/HTTP/Status/101
     """
-    network_backend = MockBackend(
+    network_backend = httpcore.MockBackend(
         [
             b"HTTP/1.1 101 Switching Protocols\r\n",
             b"Connection: upgrade\r\n",
@@ -591,7 +767,7 @@ def test_http11_upgrade_connection():
             b"...",
         ]
     )
-    with ConnectionPool(
+    with httpcore.ConnectionPool(
         network_backend=network_backend, max_connections=1
     ) as pool:
         with pool.stream(
